from __future__ import annotations
from typing import Generator, Any

import requests

from . import exceptions, objects

PYPI_HOST = "https://pypi.org" 
SIMPLE_CONTENT_TYPE = "application/vnd.pypi.simple.v1+json"
USER_AGENT = "pypiwrap/1.0.0 (github: aescarias)"

#? The constants below are marked for potential deprecation.
JSON_URL = f"{PYPI_HOST}/pypi"
STATS_URL = f"{PYPI_HOST}/stats"
SIMPLE_URL = f"{PYPI_HOST}/simple"


class PyPIClient:
    """Client for the PyPI JSON and Stats API"""

    def __init__(self, host=PYPI_HOST) -> None:
        self.host = host
        self.rest = requests.Session()
        self.rest.headers["User-Agent"] = USER_AGENT

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.rest.close()

    def get_project(self, name: str, version: str | None = None) -> objects.Project:
        """Gets information about a project or any of its releases.
        
        Arguments:
            name (:class:`str`): The name of the project
            
            version (:class:`str`, optional):
                A version of the project to fetch. If none specified,
                the latest will be fetched.
        """
        
        if version:
            rs = self.rest.get(f"{self.host}/pypi/{name}/{version}/json")
        else:
            rs = self.rest.get(f"{self.host}/pypi/{name}/json")
        
        exceptions.raise_for_status(rs, {
            404: f"Could not find project or release for '{name}'"
        })
        
        return objects.Project._from_raw(rs.json())

    def get_stats(self) -> objects.Stats:
        """Gets statistics about PyPI"""
        
        rs = self.rest.get(f"{self.host}/stats", 
                           headers={ "Accept": "application/json"})
        exceptions.raise_for_status(rs)

        return objects.Stats._from_raw(rs.json())


class SimpleClient:
    """Client for the PyPI Simple API"""

    def __init__(self, host: str = PYPI_HOST) -> None:
        self.host = host
        self.rest = requests.Session()
        self.rest.headers["Accept"] = SIMPLE_CONTENT_TYPE
        self.rest.headers["User-Agent"] = USER_AGENT

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.rest.close()
    
    def get_index(self) -> Generator[str, None, None]:
        """Yields a list of names for all projects registered on this repository"""
        rs = self.rest.get(f"{self.host}/simple")
        exceptions.raise_for_status(rs)

        for project in rs.json()["projects"]:
            yield project["name"]

    def get_page(self, project: str) -> objects.ProjectPage:
        """Gets the page for a given project"""
        rs = self.rest.get(f"{self.host}/simple/{project}")
        exceptions.raise_for_status(rs)
        
        return objects.ProjectPage._from_raw(rs.json())



from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
from typing import Any

from .base import APIObject
from ..utils import Size, iso_to_datetime, remove_additional
    

@dataclass
class Project(APIObject):
    """A PyPI project. This includes information about the project, its releases, 
    and vulnerabilities."""

    author: str
    """The author of this project"""

    author_email: str
    """The email of the project's author"""

    bugtrack_url: str | None
    """A bug tracking URL if available
    
    .. deprecated:: 1.1.
       Will be removed in 2.0. Check for a bugtrack URL in :attr:`.project_urls` instead.
    """

    classifiers: list[str]
    """A list of PyPI [classifiers](https://pypi.org/classifiers) for this project"""

    description: str
    """A description of the project"""
    
    description_content_type: str | None
    """The content type of the description if available"""

    docs_url: str | None
    """The documentation URL for the project if available"""

    download_url: str
    """The project's download URL"""

    home_page: str
    """The project's home page"""

    keywords: str
    """Keywords relating to this project"""

    license: str
    """The license for this project"""

    maintainer: str
    """The project's maintainer"""

    maintainer_email: str
    """The email of the project's maintainer"""

    name: str
    """The name of the project"""

    package_url: str
    """The PyPI package URL. Read :attr:`.project_url` for details."""

    platform: str | None
    """The release's platform target if any specified"""

    project_urls: dict[str, str]
    """A mapping of labels to URLs relating to the project"""

    project_url: str
    """The PyPI project's URL.
    
    :attr:`.package_url` and :attr:`.project_url` are effectively the same value.
    This attribute should be preferred ([see this](https://github.com/pypi/warehouse/issues/3206))
    """

    release_url: str
    """The project URL relating to this specific release"""

    requires_dist: list[str]
    """A list of required distributions or dependencies in a format similar to a requirements file"""

    requires_python: str | None
    """The Python version required for this release"""

    summary: str
    """A short summary of the project"""

    version: str
    """The version of the project"""

    yanked: bool
    """Whether this release was 'yanked' or removed from circulation"""

    yanked_reason: str | None
    """The reason the release was yanked if applicable"""

    file_urls: list[ReleaseFile]
    """A list of files for this release"""

    vulnerabilities: list[Vulnerability]
    """A list of vulnerabilities for this release if any"""

    last_serial: int
    """This is an internal value used by PyPI to indicate last modification."""

    @classmethod
    def _from_raw(cls, data: dict) -> Project:
        info = remove_additional(cls, data["info"].copy())

        vulns = list(map(Vulnerability._from_raw, data["vulnerabilities"]))
        files = list(map(ReleaseFile._from_raw, data["urls"]))

        if not data["info"].get("requires_dist"):
            data["info"]["requires_dist"] = []

        return cls(**info,
            last_serial=data["last_serial"],
            vulnerabilities=vulns,
            file_urls=files
        )

    def __repr__(self) -> str:
        return self._build_repr(self.name, 
            summary=self.summary, 
            version=self.version, 
            package_url=self.package_url
        )


@dataclass
class Vulnerability(APIObject):
    """A vulnerability in a project or release"""

    aliases: list[str]
    """The names used to refer to this vulnerability"""

    details: str
    """Details about the vulnerability"""

    fixed_in: list[str]
    """Releases where this vulnerability was fixed"""

    id: str
    """Identifier for this vulnerability"""
    
    link: str
    """A URL where more information is provided about this vulnerability"""
    
    source: str
    """The source from where this vulnerability report was obtained"""

    summary: str | None
    """A short summary of this vulnerability if available"""

    withdrawn: datetime | None = None
    """The datetime this vulnerability was withdrawn"""

    @classmethod
    def _from_raw(cls, data: dict) -> Vulnerability:
        if data.get("withdrawn") is not None:
            data["withdrawn"] = iso_to_datetime(data["withdrawn"])

        return cls(**remove_additional(cls, data))
    
    def __repr__(self) -> str:
        return self._build_repr(id=self.id, source=self.source, withdrawn=self.withdrawn)
    

@dataclass
class ReleaseFile(APIObject):
    """A file part of a PyPI release"""

    comment_text: str
    """A comment for this release"""

    digests: dict[str, str]
    """A mapping of hashes corresponding to this release file.

    Most commonly, the digests available are ``md5``, ``sha256``, and ``blake2b_256``. 
    The keys available must be members of :attr:`hashlib.algorithms_guaranteed`.
    """
    
    # downloads: int
    # md5_digest: str

    filename: str
    """The filename for this release file"""
   
    has_sig: bool
    """Whether this release file has a PGP/GPG signature attached to it.
    
    .. deprecated:: 1.1
        Will be removed in 2.0. If you need a signature, provide the :attr:`.url` 
        attribute to :func:`pypiwrap.utils.gpg_from_url`.
    """

    package_type: str  # API: packagetype
    """The type of release file. It can be either of:
    
    - ``sdist``: A source distribution (generally a .tar.gz file)
    - ``bdist_*``: A built distribution where `*` is either `wheel` or `egg`
    """

    python_version: str
    """The general Python version target for this file.
    
    It is 'source' for source distributions and a version target for built distributions.
    """

    requires_python: str
    """The Python version constraints for this file"""

    size: Size
    """The size of the release file"""

    upload_time: datetime
    """The time this file was uploaded on"""
    
    upload_time_tz: datetime # API: upload_time_iso_8601
    """The time this file was uploaded on in UTC and compliant with ISO 8601"""

    url: str
    """The URL for this release file"""

    yanked: bool
    """Whether this package was yanked"""

    yanked_reason: str | None
    """Why the package was yanked if applicable"""

    @classmethod
    def _from_raw(cls, data: dict) -> ReleaseFile:
        data = data.copy()

        # Converting values to appropriate types
        data["size"] = Size.from_int(data["size"])
        data["upload_time_iso_8601"] = iso_to_datetime(data["upload_time_iso_8601"])
        data["upload_time"] = datetime.fromisoformat(data["upload_time"])

        # Renaming values to appropriate
        data["upload_time_tz"] = data.pop("upload_time_iso_8601")
        data["package_type"] = data.pop("packagetype")

        return cls(**remove_additional(cls, data))

    def __repr__(self) -> str:
        return self._build_repr(self.filename, 
            size=self.size.si, 
            package_type=self.package_type
        )


@dataclass
class Stats(APIObject):
    """Statistics about PyPI"""

    total_size: Size
    """The total size of all packages on PyPI combined"""
    
    top_packages: dict[str, Size]
    """Mapping of top packages sorted by size"""
    
    @classmethod
    def _from_raw(cls, data: dict[str, Any]) -> Stats:
        top_pkgs_sort = sorted(data["top_packages"].items(), 
                               key=lambda it: it[1]["size"])

        top_pkgs = { name: Size.from_int(pkg["size"]) for name, pkg in top_pkgs_sort } 
        
        return cls(
            total_size=Size.from_int(data["total_packages_size"]),
            top_packages=top_pkgs   
        )
    
    def __repr__(self) -> str:
        return self._build_repr(total_size=self.total_size.si)



from __future__ import annotations

import dataclasses
from datetime import datetime
from typing import NamedTuple, Any

import requests

SI_SUFFIXES  = ["B", "KB", "MB", "GB", "TB"]
IEC_SUFFIXES = ["B", "KiB", "MiB", "GiB", "TiB"]


class Size(NamedTuple):
    """A tuple that includes human-readable representations of a file size"""

    bytes: int
    """The size represented in bytes"""
    iec: str
    """The size represented in binary/IEC units (KiB, MiB)"""
    si: str
    """The size represented in decimal/SI units (KB, MB)"""

    @classmethod
    def from_int(cls, num: int) -> Size:
        return Size(
            num, 
            iec=bytes_to_readable(num, 'iec'), 
            si=bytes_to_readable(num, 'si')
        )

    def __int__(self) -> int:
        return self.bytes

def iso_to_datetime(iso: str) -> datetime:
    """Converts an ISO 8601 string to a datetime object"""
    try:
        return datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S.%f%z")
    except ValueError:
        return datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S%z")

def gpg_from_url(url: str) -> str | None:
    """Gets the GPG signature of a file from its URL if available"""

    rs = requests.get(url + ".asc")
    if rs.ok:
        return rs.text

# where unit is either of 'si' or 'iec'
def bytes_to_readable(num: int, unit: str = 'si') -> str:
    """Converts a number (in bytes) to a human-readable string representation
    
    Arguments:
        number (:class:`int`): A value in bytes.

        unit (:class:`str`, optional): 
            Determines how to represent the output. This is either  ``si`` 
            (for decimal [MB]; default) or ``iec`` (for binary [MiB]).
    """
    if unit == 'iec':
        suffixes = IEC_SUFFIXES
        step_unit = 1024
    else:
        suffixes = SI_SUFFIXES
        step_unit = 1000

    # last one is assumed to be greatest
    suffix = None
    for suffix in suffixes:
        if num < step_unit:
            break
        num /= step_unit # type: ignore
    
    return f"{num:.2f} {suffix or suffixes[-1]}"

def remove_additional(cls: type[Any], data: dict[str, Any]) -> dict[str, Any]:
    """Takes any dataclass and a dictionary that can unpack to it
    and discards any additional keys not part of the dataclass.
    
    Returns:
        A ``dict`` that can safely unpack to the dataclass.
    """

    result = data.copy()
    names = [field.name for field in dataclasses.fields(cls)]

    for key in data:
        if key not in names:
            result.pop(key)

    return result



from __future__ import annotations
import requests

def raise_for_status(
    rs: requests.Response, 
    messages: dict[int, str] | None = None
) -> None:
    """Raises an exception based on the response's status code.
    If the status code is deemed OK, this will do nothing.
    
    Arguments:
        rs (:class:`requests.Response`): The response itself.

        messages (:class:`dict[int, str]`, optional):
            An *optional* mapping of status codes to messages.        
    """
    if rs.ok:
        return
    
    if messages is None:
        messages = {}

    error_map = { 404: NotFound }

    exc = error_map.get(rs.status_code, ClientError)
    raise exc(rs.status_code, messages.get(rs.status_code, rs.reason))


class ClientError(Exception):
    """Raised when an error occurs while performing a request."""
    def __init__(self, status: int, reason: str) -> None:
        self.status = status
        self.reason = reason

        super().__init__(f"{status}: {reason}")


class NotFound(ClientError):
    """Raised when a project, release, or file was not found"""
    pass



from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime

from .base import APIObject
from ..utils import Size, iso_to_datetime, remove_additional


@dataclass
class DistributionFile(APIObject):
    """A file for a package distribution"""

    filename: str
    """The filename of the distribution"""

    url: str
    """The download URL for the file"""

    size: Size
    """The size of the distribution"""

    hashes: dict[str, str]
    """A mapping of hashes for this file. Similar to :attr:`.ReleaseFile.digests`."""

    upload_time: datetime | None = None
    """The upload time for this file"""

    requires_python: str | None = None
    """The version constraints for this file if specified"""

    core_metadata: bool | dict[str, str] | None = None
    """
    - If a boolean, whether this file has associated metadata.
    - If a dictionary, a mapping of hashes to encoded metadata file hashes.
    """

    dist_info_metadata: bool | dict[str, str] | None = None
    """
    Contains the same values as :attr:`DistributionFile.core_metadata`. Where available, 
    prefer the use of ``core_metadata`` over this attribute.
    """

    has_sig: bool | None = None  # API: gpg_sig
    """Whether a GPG signature is included with the file"""

    yanked: bool | str | None = None
    """
    - If a boolean, whether the file was yanked.
    - If a string, why the package was yanked.
    """

    @classmethod
    def _from_raw(cls, data: dict) -> DistributionFile:
        # Certain API attributes, like requires-python, must be converted
        # to snake_case before unpacking
        result = { k.replace("-", "_"): v for k, v in data.items() }
        
        result["has_sig"] = result.get("gpg_sig")
        result["size"] = Size.from_int(result["size"])
        result["dist_info_metadata"] = result.pop("data_dist_info_metadata")

        if result.get("upload_time") is not None:
            result["upload_time"] = iso_to_datetime(result["upload_time"])

        return cls(**remove_additional(cls, result))
        
    def __repr__(self) -> str:
        return self._build_repr(self.filename, size=self.size.si)


@dataclass
class ProjectPage(APIObject):
    """A Simple project page"""

    name: str
    """The name of this project"""
    
    versions: list[str]
    """A list of all available versions for this project"""
    
    files: list[DistributionFile]
    """A list of distribution files for this project"""

    @classmethod
    def _from_raw(cls, data: dict) -> ProjectPage:
        files = [DistributionFile._from_raw(fl) for fl in data["files"]]
        return cls(name=data["name"], versions=data["versions"], files=files)

    def __repr__(self) -> str:
        return self._build_repr(self.name)



from __future__ import annotations

from ..utils import remove_additional


class APIObject:
    """A base API object"""
    
    @classmethod
    def _from_raw(cls, data: dict) -> APIObject:
        return cls(**remove_additional(cls, data.copy()))
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _build_repr(self, *args, **kwargs) -> str:
        arg_string = " ".join(map(repr, args))
        kwarg_string = " ".join(f"{k}={repr(v)}" for k, v in kwargs.items())
        final = (self.__class__.__name__, arg_string, kwarg_string)

        return "<" + ' '.join(filter(None, final)).strip() + ">"



"""
## pypiwrap

A Python API wrapper for the Python Package Index.
"""

from .client import SimpleClient, PyPIClient

__name__ = "pypiwrap"
__author__ = 'Angel Carias'
__license__ = 'MIT'
__version__ = '1.1.0'
